#include "iniapi.h"
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <share.h>
#include <sys/types.h>
#include <sys/stat.h>

/*man
NAME
	getinientry - get an entry from ini file

SYNOPSIS
	#include "iniapi.h"

	int getinientry(FILE *fp,char *entry,char *result,long *pos)

DESCRIPTION
	find specified entry from ini file.
	assumes that the correct section is already found and starts from
	current file position
	returns 1 and result filled with entry data
	or 0 and blank result if no such entry was found
	search is ended when file ends or another section is found
*/

int getinientry(FILE *fp,char *entry,char *result,long *pos)
{
char s[512],*t;
	while (!feof(fp)) {
		if (pos)
			*pos=ftell(fp);						/* save position */
		if (fgets(s,sizeof s,fp)!=NULL) {
			t=iniparse(s,s,sizeof s);
			if (*s=='[')						/* found another section */
				break;
			if (!*s)							/* got comment */
				continue;
			if (!strcmpi(s,entry)) {		/* got match, get result */
				if (result) {
					strcpy(result,t);		/* result required */
				}
				return 1;					/* found the entry */
			}
		}
	}
	if (result)
		*result='\0';
	if (pos)
		*pos=0;
	return 0;									/* return bad... */
}



/*man
NAME
	getini - gets a value of single entry in INI file

SYNOPSIS
	#include "iniapi.h"

	int getini(char *filename,char *section,char *entry,char *result)

DESCRIPTION
	gets ini file entry under specific section of specified file
	returns 1 and result filled with data if entry is found
	returns 0 and result cleared with data
*/
int getini(char *filename,char *section,char *entry,char *result)
{
FILE *fp;
int i=0;
	if (result)
		*result='\0';
	fp=openinifile(filename,"rt");			/* open the file */
	if (fp) {
		if (findinisection(fp,section,NULL)) {	/* go to section */
			i=getinientry(fp,entry,result,NULL);/* get entry (if any) */
		}
		closeinifile(fp);							/* close file */
	}
	return i;									/* return result */
}

/*man
NAME
	findinisection - seek a given section in INI file

SYNOPSIS
	#include "iniapi.h"

	int findinisection(FILE *fp,char *section,long *pos)

DESCRIPTION
	find a given section ini file. starts from the current file position.
	if you set section to NULL, it will read thru the file until any
	section header or end of file. The position of the header line will
	be placed to *pos if the header was found. Returns 1 on success,
	0 on failure.
*/
int findinisection(FILE *fp,char *section,long *pos)
{
char s[512],*t;
long p;
int i=0;
	while (!i && !feof(fp)) {
		p=ftell(fp);
		if (fgets(s,sizeof s,fp)!=NULL) {
			iniparse(s,s,sizeof s);
			if (*s=='[') {						/* found section */
				if (!section) {
					i=1;
				}
				t=strstr(s,"]");
				if (t)							/* strip closing bracket */
					*t='\0';
				if (!strcmpi(s+1,section))
					i=1;						/* got the right one */
			}
		}
	}
	if (pos)
		*pos=p;
	return i;
}

/*man
NAME
	openinifile - opens an ini file

SYNOPSIS
	#include "iniapi.h"

	FILE *openinifile(char *fn,char *mode)

DESCRIPTION
	opens INI file like fopen, but with sharing.
*/
FILE *openinifile(char *fn,char *mode)
{
	return _fsopen(fn,mode,SH_DENYWR);
}

/*man
NAME
	closeinifile - closes an ini file

SYNOPSIS
	#include "iniapi.h"

	void closeinifile(FILE *fp)

DESCRIPTION
	closes ini file that was opened with openinifile() function
*/
void closeinifile(FILE *fp)
{
	fclose(fp);
}

/*man
NAME
	iniload - INI file loader

SYNOPSIS
	#include "iniapi.h"

	int loadinifile(char *fn,INIFIXUP *fixups)

DESCRIPTION
	this function allows you to load most of your INI file
	setting in one go. you just have to build a list of the
	entries you want to load, also specifying what type of
	the data the entry is and where to store the conversion
	result.

	the function can only handle one piece of data per entry
	and line lengths less than 256 characters and section names
	of up to 32 characters.

	you can specify multiple fixups for one INI entry. if the
	entry is not present in the ini file, then the data is not
	changed, so set your data to default values and then call
	loadinifile() to override the defaults.

	section names are run thru iniparse() so they cannot contain
	spaces.

DIAGNOSTICS
	it returns number of ini entries actually loaded.

EXAMPLE
	#include "iniapi.h"

	long i_long,i_minuslong;
	unsigned long i_longhex;
	short i_int,i_minusint;
	unsigned short i_uint;
	char yes,yea,yep,no,nope,nada;

	char path[65],string[65],qstring[65],qestring[65];

	INIFIXUP fixups[] = {
		{IF_LONG,	"numbers",	"long",		&i_long,		0 },
		{IF_ULONG,	"numbers",	"longhex",	&i_longhex,		0 },
		{IF_LONG,	"numbers",	"minuslong",&i_minuslong,	0 },
		{IF_INT,	"numbers",	"int",		&i_int,			0 },
		{IF_INT,	"numbers",	"minusint",	&i_minusint,	0 },
		{IF_UINT,	"numbers",	"uint",		&i_uint,		0 },
		{IF_BOOL,	"boolean",	"yes",		&yes,			0 },
		{IF_BOOL,	"boolean",	"yea",		&yea,			0 },
		{IF_BOOL,	"boolean",	"yep",		&yep,			0 },
		{IF_BOOL,	"boolean",	"no",		&no,			0 },
		{IF_BOOL,	"boolean",	"nope",		&nope,			0 },
		{IF_BOOL,	"boolean",	"nada",		&nada,			0 },
		{IF_PATH,	"strings",	"path",		path,	sizeof path },
		{IF_STRING,	"strings",	"string",	string,	sizeof string },
		{IF_STRING,	"strings",	"qstring",	qstring,sizeof qstring },
		{IF_ESTRING,"strings",	"qestring",	qestring,sizeof qestring },
		{IF_END,	NULL,		NULL,	NULL,	0}
	};

	void main(void)
	{
		loadinifile("demo.ini",fixups);
	}
*/
int loadinifile(char *fn,INIFIXUP *fixups)
{
FILE *fp;
char section[33],ln[256],*t,*tt;
int i,base,count=0;
long l;
unsigned long ul;
	fp=openinifile(fn,"rt");
	*section='\0';
	if (fp) {
		while (!feof(fp) && !ferror(fp)) {
			if (fgets(ln,sizeof ln,fp)) {
				t=iniparse(ln,ln,sizeof ln);
				if (*ln) {
					if (*ln=='[') {	// new section
						tt=strchr(ln,']');
						if (tt)
							*tt='\0';
						strncpy(section,ln+1,sizeof section-1);
						section[sizeof section-1]='\0';
					}
					else {			// new entry
						if (*section) {
							for (i=0;fixups[i].type!=IF_END;i++) {
								if (!strcmpi(section,fixups[i].section) &&
										!strcmpi(ln,fixups[i].entry)) {
									count++;
									switch (fixups[i].type) {
										case IF_LONG:		// signed long
										case IF_ULONG:		// unsigned long
										case IF_INT:		// signed short
										case IF_UINT:		// unsigned short
											iniparse(t,ln,sizeof ln);
											t=ln; base=10;
											if (toupper(ln[1])=='X') {
												t=ln+2; base=16;
											}
											if (fixups[i].type==IF_ULONG ||
													fixups[i].type==IF_UINT)
												ul=strtoul(t,NULL,base);
											else
												l=strtol(t,NULL,base);
											switch (fixups[i].type) {
												case IF_LONG:
													*(long *)fixups[i].data=l;
													break;
												case IF_ULONG:
													*(unsigned long*)fixups[i].data=ul;
													break;
												case IF_INT:
													*(short *)fixups[i].data=(short)l;
													break;
												case IF_UINT:
													*(unsigned short *)fixups[i].data=(unsigned short)ul;
													break;
											}
											break;
										case IF_PATH:		// pathname (backslash will be added)
											iniparse(t,fixups[i].data,fixups[i].slen);
											if (strlen(fixups[i].data)<fixups[i].slen-1) {
												tt=strchr(fixups[i].data,0)-1;
												if (*tt!='\\') {
													tt++;
													*tt++='\\';
													*tt='\0';
												}
											}
											break;
										case IF_STRING:		// string
											iniparse(t,fixups[i].data,fixups[i].slen);
											break;
										case IF_BOOL:		// 1-byte logical (Yes/No)
											iniparse(t,ln,sizeof ln);
											if (toupper(*ln)=='Y' || *ln=='1')
												*(char *)fixups[i].data=1;
											else
												*(char *)fixups[i].data=0;
											break;
										case IF_ESTRING:	// quoted/escaped string
											iniparseq(t,fixups[i].data,fixups[i].slen);
											break;
									}
								}
							}
						}
					}
				}
			}
		}
		closeinifile(fp);
	}
	return count;
}


/*man
NAME
	iniparse - parse INI file line

SYNOPSIS
	#include "iniapi.h"

	char *iniparse(char *s,char *buf,int buflen)

DESCRIPTION
	Ini line parser. Extracts components from the line and returns pointer
	to next component. If buf is NULL, skips one component. Converts
	unquoted characters to uppercase and removes quotes from quoted
	components. Can parse in place, ie.
	next=iniparse(source,source,maxlen)

SEE ALSO
	iniparseq()
*/
char *iniparse(char *s,char *buf,int buflen)
{
int quoted=0;
	if (s) {
		if (!buf)
			buflen=1;
		else
			buflen--;
		while (*s) {
			if (!quoted && (*s==',' || *s=='=')) {
				s++;
				break;
			}
			if (!quoted && isspace(*s)) {
				s++;
				continue;
			}
			if (!quoted && *s==';') {
				s=NULL;
				break;
			}
			if (*s=='\"') {
				quoted=!quoted;
				s++;
				continue;
			}
			if (buf) {
				if (buflen && *s) {
					if (quoted)
						*buf++=*s++;
					else
						*buf++=toupper(*s++);
					buflen--;
				}
				else
					s++;
			}
			else
				s++;
		}
		if (buf)
			*buf='\0';
	}
	if (s && *s)
		return s;
	else
		return NULL;
}


/*man
NAME
	iniparseq - parse INI file line with escaped characters

SYNOPSIS
	#include "iniapi.h"

	char *iniparseq(char *s,char *buf,int buflen)

DESCRIPTION
	Ini line parser. Extracts components from the line and returns
	pointer to next component. If buf is NULL, skips one component.
	Converts unquoted characters to uppercase and removes quotes
	from quotes components. Can parse in place, ie.
	next=iniparse(source,source,maxlen)

	this one treats backslash as quote character inside quotes and
	thus allows to specify characters otherwise invalid

	the specially handled characters are
	\\   - backslash itself
	\r   - CR
	\n   - NL
	\t   - TAB
	\"   - quotation marks
	\123 - byte value in decimal
	\x1f - byte value in hex

	any other character is treated like \\ (only the character
	is copied)

SEE ALSO
	iniparse()
*/
char *iniparseq(char *s,char *buf,int buflen)
{
int state='C',c,val;
	if (s) {
		if (!buf)
			buflen=1;
		else
			buflen--;
		while (s && *s) {
			c=*s;
			switch (state) {
				case 'C':		// copy with conversion
					if (c=='"') {
						state='c';
						s++;
						break;
					}
					if (c==',' || c=='=') {
						s++;
						goto getout;
					}
					if (c==';') {
						s=NULL;
						goto getout;
					}
					if (isspace(c)) {
						s++;
						break;
					}
					c=toupper(c);
store:
					if (buf) {
						if (buflen) {
							*buf++=c;
							buflen--;
							s++;
						}
						else
							s++;
					}
					else
						s++;
					break;
				case 'c':		// copy and don't convert case
					if (c=='\\') {
						state='E';
						s++;
						break;
					}
					if (*s=='"') {
						state='C';
						s++;
						break;
					}
					goto store;
				case 'E':		// escaped
					if (c=='x' || c=='X') {
						state='X';
						s++;
						break;
					}
					if (isdigit(c)) {
						state='0';
						s++;
						val=c-'0';
						break;
					}
					state='c';
					if (c=='r')
						c='\r';
					if (c=='n')
						c='\n';
					if (c=='t')
						c='\t';
					goto store;
				case '0':		// parsing decimal
					if (!isdigit(c)) {
						state='c';
						s--;
						c=val;
						goto store;
					}
					val=val*10+(c-'0');
					s++;
					break;
				case 'X':		// parsing hex
					if (!isxdigit(c)) {
						state='c';
						s--;
						c=val;
						goto store;
					}
					c=toupper(c)-'0';
					if (c>9)
						c=c-7;
					val=val*16+c;
					s++;
					break;
			}
		}
	}
getout:
	if (buf)
		*buf='\0';
	if (s && *s)
		return s;
	else
		return NULL;
}



